local L = LibStub("AceLocale-3.0"):GetLocale("Turninator")

--==== set upvalues ====--
local _G, Turninator = _G, Turninator
local format, ipairs, pairs, rawset, setmetatable, tinsert, tsort, type = format, ipairs, pairs, rawset, setmetatable, table.insert, table.sort, type

-- returns the "static" instance of a class supporting base values and/or getter-setter backed properties
local function NewGSBaseClass(defaults, getters, setters)
	-- these locals should continue to exist in a closure after function return
	local newclass = defaults or {}
	local newclass_mt = {}

	newclass_mt.__index = getters and function(t, k)
			local get = getters[k]
			local retval = nil

			if (type(get) == "function") then
				retval = get(t,k)
			elseif (type(get) == "table") then
				retval = get[k]
			end

			if (retval == nil) then
				retval = newclass[k]
			end

			return retval
		end or newclass

	newclass_mt.__newindex = setters and function(t, k, v)
			local set = setters[k]
			if (set == nil) then
				rawset(t,k,v)
			elseif (type(set) == "table") then
				set[k] = v
			elseif (type(set) == "function") then
				set(t,k,v)
			end
		end


	-- returns an instance of class
	function newclass:New(newinstance)
		newinstance = newinstance or {}
		return setmetatable(newinstance, newclass_mt)
	end

	return newclass
end


local ITEM = NewGSBaseClass(
	{	-- defaults
		icon = "Interface\\Icons\\INV_Misc_QuestionMark"
	},
	{	-- getters
		name = function(t)
			-- print("name calling", t.__npcname, Turninator.db.global.tblUseItem[t.__npcname])
			return Turninator.db.global.tblUseItem[t.__npcname]
		end,
		icon = function(t)
			return Turninator.db.global.tblItemIcon[t.name]
		end,
	},
	{	-- setters
		name = function(t,_,v)
			-- print("name calling", t.__npcname, Turninator.db.global.tblUseItem[t.__npcname])
			Turninator.db.global.tblUseItem[t.__npcname] = v
		end,
	}
)

local NPC = NewGSBaseClass( {},
	{	-- getters
		interact = function(t)
			-- print("name calling", t.__npcname, Turninator.db.global.tblUseItem[t.__npcname])
			return Turninator.db.global.tblInteract[t.name]
		end,
	},
	{	-- setters
		interact = function(t,_,v)
			Turninator.db.global.tblInteract[t.name] = v
		end,
	},
	function(t) t.item = ITEM:New( { __npcname = t.name } ) end
)
do
	local origNew = NPC.New
	function NPC:New(name)
		local npc = origNew( NPC, { name = name } )
		npc.item = ITEM:New( { __npcname = name } )
		return npc
	end
end

function NPC:GetNameColor()
	if (self.interact == nil) then
		return "FFFFFF00"	-- yellow
	elseif (self.interact == true) then
		return "FF00FF19"	-- green
	elseif (self.interact == false) then
		return "FFD82619"	-- red
	end
end

function NPC:GetShortDisplayName()
	return format("|T%s:0|t  |c%s%s|r", self.item.name and self.item.icon or "", self:GetNameColor(), self.name)
end

function NPC:GetLongDisplayName()
	if self.item.name then
		return format("|c%s%s|r     |T%s:0|t%s", self:GetNameColor(), self.name, self.item.icon, self.item.name)
	else
		return format("|c%s%s|r", self:GetNameColor(), self.name)
	end
end


NPC.options = {
	item = {
		type = "description",
		order = 1,
		width = "full",
		fontSize = "large",
		image = function(info) return info.handler.item.icon end,
		hidden = function(info) return not info.handler.item.name end,
		name = function(info) return info.handler.item.name or "" end,
	},
	interact = {
		type = "toggle",
		order = 2,
		width = "full",
		name = L["Interact with NPC"],
		hidden = function(info) return info.handler.interact == nil end,
		get = function(info) return info.handler.interact end,
		set = function(info,value) info.handler.interact = value end,
	},
	remove = {
		type = "execute",
		order = 3,
		name = L["Remove"],
		desc = L["Removes the NPC and any associated interations"],
		func = function(info) 
				local name = info.handler.name
				Turninator.npcs[name] = nil
				Turninator.db.global.tblUseItem[name] = nil
				Turninator.db.global.tblInteract[name] = nil
				LibStub("AceConfigRegistry-3.0"):NotifyChange("Turninator");
			end,
	},
}

function NPC.GetOptions()
	local names = {}
	for k,_ in pairs(Turninator.db.global.tblUseItem) do
		tinsert(names, k)
	end
	for k,_ in pairs(Turninator.db.global.tblInteract) do
		if (not Turninator.db.global.tblUseItem[k]) then
			tinsert(names, k)
		end
	end
	tsort(names)

	Turninator.npcs = {}
	for order,name in ipairs(names) do
		local npc = NPC:New(name)
		Turninator.npcs[name] = {
			name = function(info) return info.handler:GetShortDisplayName() end,
			order = order,
			type = 'group',
			inline = false,
			handler = npc,
			args = NPC.options,
		}
	end
	return Turninator.npcs
end

function Turninator:LoadSavedSettings()
	-- saved settings and default values
	local defaults = {
		global = {
			iDebugLevel = 0,
			tblInteract = {},   -- table { key(string) name of npc,  value(boolean)
			tblUseItem  = {},   -- table { key(string) name of npc,  value(string) name of item to use
			tblItemIcon = {},   -- table { key(string) name of item, value(string) icon texture path
		},
		profile = {
			iDataFormat = 0,
			fEnabled = true,
			fDisableInInstances = false,
		},
	}

	self.db = LibStub("AceDB-3.0"):New("TurninatorDB", defaults)
end


--=========================================================================--
-- Options UI definition
--=========================================================================--

function Turninator.GetUIOptions(uiType, uiName, appName)
--	print("GetUIOptions   ", uiType)
	
	if (not Turninator.options) then
		Turninator.options = {}
		
		-- start options.commands --- commands executable from command line
		Turninator.options.commands = {
			type = 'group',
			name = "Turninator",
			handler = Turninator,
			args = {
				Debug = {
					hidden = function(info) return Turninator.db.global.iDebugLevel == 0 end,
					order = 0,
					type = "input",
					name = "debug",
					desc = "Sets verbosity of debug output",
					get = function(info) return Turninator.db.global.iDebugLevel end,
					set = "Debug",
				},
				SetIcon = {
					hidden = true,
					order = 1,
					type = "input",
					name = "seticon",
					desc = "caches the icon texture path for a given item id",
					get = function(info) return 0 end,
					set = "SetIconById",
				},
				[ L["On"] ] = {
					order = 5,
					type = "execute",
					name = L["On"],
					desc = L["Enables the addon"],
					func = "SetEnabledState",
				},
				[ L["Off"] ] = {
					order = 10,
					type = "execute",
					name = L["Off"],
					desc = L["Disables the addon"],
					func = "SetEnabledState",
				},
				[ L["Config"] ] = {
					order = 15,
					type = "execute",
					name = L["Config"],
					desc = L["Display options UI"],
					-- func = function(info) InterfaceOptionsFrame_OpenToCategory(Turninator.optionsFrame) end,
					func = "Config",
				},
				[ L["ShareKey"] ] = {
					hidden = true,
					order = 20,
					type = "input",
					name = L["ShareKey"],
					desc = L["Sets the keybind to use with the addon"],
					get = function(info) return Turninator.db.profile.strSharedKey end,
					set = "ShareKey",
				},
				[ L["ShareCmd"] ] = {
					hidden = true,
					order = 25,
					type = "input",
					name = L["Shared Command is:"],
					desc = L["Sets the addon to share the keybind set for that command"],
					get = function(info) return Turninator.db.profile.strSharedCmd end,
					set = "ShareCmd",
				},
				[ L["Reset"] ] = {
					hidden = true,
					order = 30,
					type = "execute",
					name = L["Reset"],
					desc = L["Decouples the addon from the shared keybind"],
					func = "Reset",
				},
				[ L["UseItem"] ] = {
					order = 35,
					type = "input",
					name = L["UseItem"],
					desc = L["Set binding to always use named item on targeted npc"],
					get = function(info) return "" end,
					set = "UseItem",
				},
				[ L["Interact"] ] = {
					order = 25,
					type = "execute",
					name = L["Interact"],
					desc = L["Treat targeted npc as interactable from now on"],
					func = "SetNPC",
				},
				[ L["Ignore"] ] = {
					order = 30,
					type = "execute",
					name = L["Ignore"],
					desc = L["Treat targeted npc as not being interactable from now on"],
					func = "SetNPC",
				},
				[ L["Forget"] ] = {
					order = 40,
					type = "execute",
					name = L["Forget"],
					desc = L["Removes targeted npc from UseItem, Interact and/or Ignore lists"],
					func = "SetNPC",
				},
				List = {
					hidden = true,
					order = 45,
					type = "execute",
					name = "List",
					desc = "Prints out the current contents of the UseItem, Interact and/or Ignore lists",
					func = "ListNPCs",
				},
			},
		}  -- end options.commands --- commands executable from command line


		-- start options.gui --- controls to display in option guis
		Turninator.options.gui = {
			type = 'group',
			name = "Turninator",
			handler = Turninator,
			childGroups = 'tree',
			inline = false,
			plugins = { People = NPC.GetOptions() },
			args = {
	--[==[
				Debug = {
					hidden = function(info) return Turninator.db.global.iDebugLevel == 0 end,
					order = 0,
					type = "select",
					name = "Debug",
					desc = "Sets verbosity of debug output",
					values = Turninator.DL,
					get = function(info) return Turninator.db.global.iDebugLevel end,
					set = "Debug",
				},
	--]==]
				[ L["ShareKey"] ] = {
					order = 20,
					type = "keybinding",
					name = L["Shared Key is:"],
					desc = L["Sets the keybind to use with the addon"],
					get = function(info) return Turninator.db.profile.strSharedKey end,
					set = "ShareKey",
				},

				PaddingShareKey = {
					order = 25,
					type = "description",
					width = "half",
					name = "",
				},

				[ L["ShareCmd"] ] = {
					order = 40,
					type = "input",
					disabled = true,
					width = "double",
					name = L["Shared Command is:"],
					desc = L["Sets the addon to share the keybind set for that command"],
					get = function(info) return Turninator.db.profile.strSharedCmd end,
					set = "ShareCmd",
				},

				PaddingShareCmd = {
					order = 45,
					type = "description",
					name = "",
				},

				Enabled = {
					order = 50,
					type = "toggle",
					name = L["Addon enabled"],
					get = function(info) return Turninator.db.profile.fEnabled end,
					set = function(info, value) Turninator:SetEnabledState(info, value and L["On"] or L["Off"]) end,
				},
--[===[
				PaddingShareKey = {
					order = 55,
					type = "description",
					width = "half",
					name = "",
				},

				DisableInInstances = {
					order = 60,
					type = "toggle",
					name = L["Disable inside instances"],
					get = function(info) return Turninator.db.profile.fDisableInInstances end,
					set = "SetDisableInInstance",
				},
--]===]
				PaddingEmptyLine = {
					order = 65,
					type = "description",
					width = "full",
					name = " ",
				},

				NPCDesc = {
					order = 70,
					type = "header",
					name = L["NPC Interactions"],
				},
			},
		}  -- end options.commands --- commands executable from command line
	end

	if (uiType == "cmd") then
		return Turninator.options.commands
	else
		return Turninator.options.gui;
	end
end
